part of maps;

/// Customizes the appearance of the selected shape.
///
/// ```dart
///   @override
///  Widget build(BuildContext context) {
///    return SfMaps(
///      layers: [
///        MapShapeLayer(
///          enableSelection: true,
///          selectionSettings: MapSelectionSettings(
///              color: Colors.black
///          ),
///          delegate: MapShapeLayerDelegate(
///              shapeFile: "assets/world_map.json",
///              shapeDataField: "name",
///              dataCount: bubbleData.length,
///              primaryValueMapper: (index) {
///                return bubbleData[index].country;
///              }),
///        )
///      ],
///    );
///  }
/// ```
class MapSelectionSettings {
  /// Creates a [MapSelectionSettings].
  const MapSelectionSettings(
      {this.color, this.strokeColor, this.strokeWidth, this.opacity = 1.0});

  /// Fills the selected shape by this color.
  ///
  /// This snippet shows how to set selection color in [SfMaps].
  ///
  /// ```dart
  ///   @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableSelection: true,
  ///          selectionSettings: MapSelectionSettings(
  ///              color: Colors.black
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "name",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  /// See also:
  /// * [strokeColor], to set stroke color for selected shape.
  final Color color;

  /// Applies stroke color for the selected shape.
  ///
  /// This snippet shows how to set stroke color for the selected shape.
  ///
  /// ```dart
  ///   @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableSelection: true,
  ///          selectionSettings: MapSelectionSettings(
  ///              strokeColor: Colors.white
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "name",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  /// See also:
  /// * [Color], to set selected shape color.
  final Color strokeColor;

  /// Stroke width which applies to the selected shape.
  ///
  /// This snippet shows how to set stroke width for the selected shape.
  ///
  /// ```dart
  ///   @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableSelection: true,
  ///          selectionSettings: MapSelectionSettings(
  ///              strokeWidth: 2
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "name",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  final double strokeWidth;

  /// Specifies the opacity of the selected shape.
  ///
  /// This snippet shows how to set opacity of the selected shape.
  ///
  /// ```dart
  ///   @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableSelection: true,
  ///          selectionSettings: MapSelectionSettings(
  ///              opacity: 0.5
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "name",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  final double opacity;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != runtimeType) {
      return false;
    }
    return other is MapSelectionSettings &&
        other.color == color &&
        other.strokeWidth == strokeWidth &&
        other.strokeColor == strokeColor &&
        other.opacity == opacity;
  }

  @override
  int get hashCode => hashValues(color, strokeWidth, strokeColor, opacity);
}
