part of maps;

/// Customizes the appearance of the bubble's or shape's tooltip.
///
/// ```dart
///  @override
///  Widget build(BuildContext context) {
///    return SfMaps(
///      layers: [
///        MapShapeLayer(
///          enableShapeTooltip: true,
///          tooltipSettings: MapTooltipSettings(
///              color: Colors.black
///          ),
///          delegate: MapShapeLayerDelegate(
///              shapeFile: "assets/world_map.json",
///              shapeDataField: "continent",
///              dataCount: bubbleData.length,
///              primaryValueMapper: (index) {
///                return bubbleData[index].country;
///              }),
///        )
///      ],
///    );
///  }
/// ```
class MapTooltipSettings {
  /// Creates a [MapTooltipSettings].
  const MapTooltipSettings({
    this.color,
    this.textStyle,
    this.strokeWidth,
    this.strokeColor,
  });

  /// Fills the tooltip by this color.
  ///
  /// This snippet shows how to set the tooltip color in [SfMaps].
  ///
  /// ```dart
  ///  @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableShapeTooltip: true,
  ///          tooltipSettings: MapTooltipSettings(
  ///              color: Colors.black
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "continent",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  /// See also:
  /// * [textStyle], for customizing the style of the tooltip text.
  final Color color;

  /// Customizes the tooltip text's style.
  ///
  /// This snippet shows how to customize the tooltip text style in [SfMaps].
  ///
  /// ```dart
  ///  @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableShapeTooltip: true,
  ///          tooltipSettings: MapTooltipSettings(
  ///              textStyle: TextStyle(color: Colors.red)
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "continent",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  final TextStyle textStyle;

  /// Specifies the stroke width applies to the tooltip.
  ///
  /// This snippet shows how to customize the stroke width in [SfMaps].
  ///
  /// ```dart
  ///  @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableShapeTooltip: true,
  ///          tooltipSettings: MapTooltipSettings(
  ///              strokeWidth: 2
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "continent",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  /// See also:
  /// * [strokeColor], for customizing the stroke color of the tooltip.
  final double strokeWidth;

  /// Specifies the stroke color applies to the tooltip.
  ///
  /// This snippet shows how to customize stroke color in [SfMaps].
  ///
  /// ```dart
  ///  @override
  ///  Widget build(BuildContext context) {
  ///    return SfMaps(
  ///      layers: [
  ///        MapShapeLayer(
  ///          enableShapeTooltip: true,
  ///          tooltipSettings: MapTooltipSettings(
  ///              strokeColor: Colors.white
  ///          ),
  ///          delegate: MapShapeLayerDelegate(
  ///              shapeFile: "assets/world_map.json",
  ///              shapeDataField: "continent",
  ///              dataCount: bubbleData.length,
  ///              primaryValueMapper: (index) {
  ///                return bubbleData[index].country;
  ///              }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  ///
  /// See also:
  /// * [strokeWidth] for customizing the stroke width of the tooltip.
  final Color strokeColor;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != runtimeType) {
      return false;
    }
    return other is MapTooltipSettings &&
        other.color == color &&
        other.strokeWidth == strokeWidth &&
        other.strokeColor == strokeColor &&
        other.textStyle == textStyle;
  }

  @override
  int get hashCode => hashValues(color, strokeWidth, strokeColor, textStyle);
}
