part of maps;

/// Base class for map icon shapes.
class _MapIconShape {
  const _MapIconShape();

  /// Returns the size based on the value passed to it.
  Size getPreferredSize(Size iconSize, SfMapsThemeData mapsThemeData) {
    return iconSize;
  }

  /// Paints the shapes based on the value passed to it.
  void paint(
    PaintingContext context,
    Offset offset, {
    RenderBox parentBox,
    SfMapsThemeData mapThemeData,
    Size iconSize,
    Color color,
    Color strokeColor,
    double strokeWidth,
    MapIconType iconType,
  }) {
    iconSize = getPreferredSize(iconSize, mapThemeData);
    final double halfIconWidth = iconSize.width / 2;
    final double halfIconHeight = iconSize.height / 2;
    final bool hasStroke =
        strokeWidth != null && strokeWidth > 0 && strokeColor != null;
    final Paint paint = Paint()
      ..isAntiAlias = true
      ..color = color;
    Path path;

    switch (iconType) {
      case MapIconType.circle:
        final double radius = max(halfIconWidth, halfIconHeight);
        context.canvas.drawCircle(offset, radius, paint);
        if (hasStroke) {
          paint
            ..strokeWidth = strokeWidth
            ..color = strokeColor
            ..style = PaintingStyle.stroke;
          context.canvas.drawCircle(offset, radius, paint);
        }
        break;
      case MapIconType.square:
        final Rect rect = Rect.fromCircle(
            center: offset, radius: max(halfIconWidth, halfIconHeight));
        context.canvas.drawRect(rect, paint);
        if (hasStroke) {
          paint
            ..strokeWidth = strokeWidth
            ..color = strokeColor
            ..style = PaintingStyle.stroke;
          context.canvas.drawRect(rect, paint);
        }
        break;
      case MapIconType.triangle:
        path = Path()
          ..moveTo(offset.dx - halfIconWidth, offset.dy + halfIconHeight)
          ..lineTo(offset.dx + halfIconWidth, offset.dy + halfIconHeight)
          ..lineTo(offset.dx, offset.dy - halfIconHeight)
          ..close();
        context.canvas.drawPath(path, paint);
        if (hasStroke) {
          paint
            ..strokeWidth = strokeWidth
            ..color = strokeColor
            ..style = PaintingStyle.stroke;
          context.canvas.drawPath(path, paint);
        }
        break;
      case MapIconType.diamond:
        path = Path()
          ..moveTo(offset.dx - halfIconWidth, offset.dy)
          ..lineTo(offset.dx, offset.dy - halfIconHeight)
          ..lineTo(offset.dx + halfIconWidth, offset.dy)
          ..lineTo(offset.dx, offset.dy + halfIconHeight)
          ..close();
        context.canvas.drawPath(path, paint);
        if (hasStroke) {
          paint
            ..strokeWidth = strokeWidth
            ..color = strokeColor
            ..style = PaintingStyle.stroke;
          context.canvas.drawPath(path, paint);
        }
        break;
    }
  }
}

/// Base class for map tooltip shapes.
class _TooltipShape {
  const _TooltipShape();

  /// Paints the tooltip shapes based on the value passed to it.
  void paint(
      PaintingContext context,
      Offset center,
      TextPainter labelPainter,
      Paint paint,
      RenderBox parentBox,
      Animation<double> tooltipAnimation,
      SfMapsThemeData themeData,
      MapTooltipSettings tooltipSettings) {
    const double _tooltipTriangleHeight = 7;
    const double cornerRadius = 4.0;
    const double _tooltipTriangleWidth = 12.0;
    const double marginSpace = 6.0;
    const double textPadding = 14.0;
    const double halfTooltipTriangleWidth = _tooltipTriangleWidth / 2;
    const double elevation = 0.0;

    Path path = Path();
    final double tooltipWidth = labelPainter.width + textPadding;
    final double tooltipHeight = labelPainter.height + textPadding;
    final double halfTooltipWidth = tooltipWidth / 2;
    final double halfTooltipHeight = tooltipHeight / 2;

    final double tooltipStartPoint = _tooltipTriangleHeight + tooltipHeight / 2;
    final double tooltipTriangleHeight =
        tooltipStartPoint - _tooltipTriangleHeight;

    final double endGlobal = parentBox.size.width - marginSpace;
    double rightLineWidth = center.dx + halfTooltipWidth > endGlobal
        ? endGlobal - center.dx
        : halfTooltipWidth;
    final double leftLineWidth = center.dx - halfTooltipWidth < marginSpace
        ? center.dx - marginSpace
        : tooltipWidth - rightLineWidth;
    rightLineWidth = leftLineWidth < halfTooltipWidth
        ? halfTooltipWidth - leftLineWidth + rightLineWidth
        : rightLineWidth;

    double moveNosePoint = leftLineWidth < tooltipWidth * 0.2
        ? tooltipWidth * 0.2 - leftLineWidth
        : 0.0;
    moveNosePoint = rightLineWidth < tooltipWidth * 0.2
        ? -(tooltipWidth * 0.2 - rightLineWidth)
        : moveNosePoint;

    final double labelOffsetY =
        _tooltipTriangleHeight + halfTooltipHeight + labelPainter.height / 2;
    double shiftText = leftLineWidth > rightLineWidth
        ? -(halfTooltipWidth - rightLineWidth)
        : 0.0;
    shiftText = leftLineWidth < rightLineWidth
        ? (halfTooltipWidth - leftLineWidth)
        : shiftText;

    rightLineWidth = rightLineWidth + elevation;

    context.canvas.save();
    context.canvas.translate(
        center.dx, center.dy - _tooltipTriangleHeight - halfTooltipHeight);
    context.canvas.scale(tooltipAnimation.value);

    path = _getTooltipPath(
        path,
        _tooltipTriangleHeight,
        halfTooltipHeight,
        halfTooltipTriangleWidth,
        tooltipTriangleHeight,
        moveNosePoint,
        rightLineWidth,
        leftLineWidth,
        cornerRadius,
        tooltipHeight);

    context.canvas.drawPath(path, paint);
    labelPainter.paint(
        context.canvas,
        Offset(-labelPainter.width / 2 + shiftText,
            tooltipStartPoint - labelOffsetY));

    final Color strokeColor =
        tooltipSettings.strokeColor ?? themeData.tooltipStrokeColor;
    if (strokeColor != null) {
      final Paint strokePaint = Paint();
      strokePaint.color = strokeColor;
      strokePaint.strokeWidth =
          tooltipSettings.strokeWidth ?? themeData.tooltipStrokeWidth;
      strokePaint.style = PaintingStyle.stroke;
      context.canvas.drawPath(path, strokePaint);
    }

    context.canvas.restore();
  }

  Path _getTooltipPath(
      Path path,
      double _tooltipTriangleHeight,
      double halfTooltipHeight,
      double halfTooltipTriangleWidth,
      double tooltipTriangleHeight,
      double moveNosePoint,
      double rightLineWidth,
      double leftLineWidth,
      double cornerRadius,
      double tooltipHeight) {
    path.reset();
    path.moveTo(0, _tooltipTriangleHeight + halfTooltipHeight);
    //    /
    path.lineTo(
        halfTooltipTriangleWidth + moveNosePoint, tooltipTriangleHeight);
    //      ___
    //     /
    path.lineTo(rightLineWidth - cornerRadius, tooltipTriangleHeight);
    //      ___|
    //     /
    path.quadraticBezierTo(rightLineWidth, tooltipTriangleHeight,
        rightLineWidth, tooltipTriangleHeight - cornerRadius);
    path.lineTo(
        rightLineWidth, tooltipTriangleHeight - tooltipHeight + cornerRadius);
    //  _______
    //      ___|
    //     /
    path.quadraticBezierTo(
        rightLineWidth,
        tooltipTriangleHeight - tooltipHeight,
        rightLineWidth - cornerRadius,
        tooltipTriangleHeight - tooltipHeight);
    path.lineTo(
        -leftLineWidth + cornerRadius, tooltipTriangleHeight - tooltipHeight);
    //  _______
    // |    ___|
    //     /
    path.quadraticBezierTo(
        -leftLineWidth,
        tooltipTriangleHeight - tooltipHeight,
        -leftLineWidth,
        tooltipTriangleHeight - tooltipHeight + cornerRadius);
    path.lineTo(-leftLineWidth, tooltipTriangleHeight - cornerRadius);
    //  ________
    // |___  ___|
    //      /
    path.quadraticBezierTo(-leftLineWidth, tooltipTriangleHeight,
        -leftLineWidth + cornerRadius, tooltipTriangleHeight);
    path.lineTo(
        -halfTooltipTriangleWidth + moveNosePoint, tooltipTriangleHeight);
    //  ________
    // |___  ___|
    //     \/
    path.close();

    return path;
  }
}
