part of maps;

///Base class for the [MapShapeLayer].
///
/// See also:
/// * [MapShapeLayer], for adding in [SfMaps.layers].
abstract class MapLayer extends StatelessWidget {
  /// Creates a [MapLayer].
  const MapLayer({Key key, this.initialMarkersCount, this.markerBuilder})
      : super(key: key);

  /// Option to set markers count initially. It cannot be be updated
  /// dynamically.
  ///
  /// The [MapLayer.markerBuilder] callback will be called number of times equal
  /// to the value specified in the [MapLayer.initialMarkersCount] property.
  /// The default value of the of this property is null.
  ///
  /// See also:
  /// * [markerBuilder], for returning the [MapMarker].
  /// * [MapShapeLayer.controller], for dynamically updating the markers
  /// collection.
  final int initialMarkersCount;

  /// Returns the [MapMarker] for the given index.
  ///
  /// Markers which be used to denote the locations on the map.
  ///
  /// It is possible to use the built-in symbols or display a custom widget at a
  /// specific latitude and longitude on a map.
  ///
  /// The [MapLayer.markerBuilder] callback will be called number of times equal
  /// to the value specified in the [MapLayer.initialMarkersCount] property.
  /// The default value of the of this property is null.
  ///
  /// For rendering the custom widget for the marker, pass the required widget
  /// for child in [MapMarker] constructor.
  ///
  /// ```dart
  /// List<Model> data;
  ///
  /// @override
  /// void initState() {
  ///    data = const <Model>[
  ///      Model('Brazil', -14.235004, -51.92528),
  ///      Model('Germany', 51.16569, 10.451526),
  ///      Model('Australia', -25.274398, 133.775136),
  ///      Model('India', 20.593684, 78.96288),
  ///      Model('Russia', 61.52401, 105.318756)
  ///    ];
  ///
  ///    super.initState();
  /// }
  ///
  ///  @override
  /// Widget build(BuildContext context) {
  ///    return Scaffold(
  ///      body: Center(
  ///          child: Container(
  ///            height: 350,
  ///            child: Padding(
  ///              padding: EdgeInsets.only(left: 15, right: 15),
  ///              child: SfMaps(
  ///                layers: <MapLayer>[
  ///                  MapShapeLayer(
  ///                    delegate: MapShapeLayerDelegate(
  ///                      shapeFile: 'assets/world_map.json',
  ///                      shapeDataField: 'name',
  ///                      dataCount: data.length,
  ///                      primaryValueMapper: (index) => data[index].country,
  ///                    ),
  ///                    initialMarkersCount: 5,
  ///                    markerBuilder: (BuildContext context, int index){
  ///                      return MapMarker(
  ///                        latitude: data[index].latitude,
  ///                        longitude: data[index].longitude,
  ///                      );
  ///                    },
  ///                  ),
  ///                ],
  ///              ),
  ///            ),
  ///          )
  ///      ),
  ///    );
  ///  }
  ///
  /// class Model {
  ///  const Model(this.country, this.latitude, this.longitude);
  ///
  ///  final String country;
  ///  final double latitude;
  ///  final double longitude;
  /// }
  /// ```
  /// See also:
  /// * [MapShapeLayerController], for dynamically updating the markers.
  /// * [MapMarker], to create a map marker.
  final MapMarkerBuilder markerBuilder;
}

/// The shape layer in which geographical rendering is done.
///
/// The actual geographical rendering is done here using the
/// [MapShapeLayer.delegate]. The path of the .json file which contains the
/// GeoJSON data has to be set to the [MapShapeLayerDelegate.shapeFile].
///
/// The [MapShapeLayerDelegate.shapeDataField] property is used to
/// refer the unique field name in the .json file to identify each shapes and
/// map with the respective data in the data source.
///
/// By default, the value specified for the
/// [MapShapeLayerDelegate.shapeDataField] in the GeoJSON file will be used in
/// the elements like data labels, tooltip, and legend for their respective
/// shapes.
///
/// However, it is possible to keep a data source and customize these elements
/// based on the requirement. The value of the
/// [MapShapeLayerDelegate.shapeDataField] will be used to map with the
/// respective data returned in [MapShapeLayerDelegate.primaryValueMapper]
/// from the data source.
///
/// Once the above mapping is done, you can customize the elements using the
/// APIs like [MapShapeLayerDelegate.dataLabelMapper],
/// [MapShapeLayerDelegate.shapeColorMappers],
/// [MapShapeLayerDelegate.shapeTooltipTextMapper], etc.
///
/// The snippet below shows how to render the basic world map using the data
/// from .json file.
///
/// ```dart
///   @override
///  Widget build(BuildContext context) {
///    return SfMaps(
///      layers: [
///        MapShapeLayer(
///          delegate: MapShapeLayerDelegate(
///              shapeFile: "assets/world_map.json",
///              shapeDataField: "name",
///           ),
///        )
///      ],
///    );
///  }
/// ```
/// See also:
/// * [delegate], to provide data for the elements of the [SfMaps] like data
/// labels, bubbles, tooltip, shape colors, and legend.
class MapShapeLayer extends MapLayer {
  /// Creates a [MapShapeLayer].
  const MapShapeLayer({
    Key key,
    @required this.delegate,
    this.loadingBuilder,
    this.controller,
    int initialMarkersCount = 0,
    MapMarkerBuilder markerBuilder,
    this.showLegend = false,
    this.showDataLabels = false,
    this.showBubbles = false,
    this.enableShapeTooltip = false,
    this.enableBubbleTooltip = false,
    this.enableSelection = false,
    this.color,
    this.strokeColor,
    this.strokeWidth,
    this.palette,
    this.legendSettings = const MapLegendSettings(),
    this.dataLabelSettings = const MapDataLabelSettings(),
    this.bubbleSettings = const MapBubbleSettings(),
    this.selectionSettings = const MapSelectionSettings(),
    this.tooltipSettings = const MapTooltipSettings(),
    this.initialSelectedIndex = -1,
    this.onSelectionChanged,
  }) : super(
          key: key,
          initialMarkersCount: initialMarkersCount,
          markerBuilder: markerBuilder,
        );

  /// The delegate that maps the data source with the shape file and provides
  /// data for the elements of the [SfMaps] like data labels, bubbles, tooltip,
  /// and shape colors.
  ///
  /// The path of the .json file which contains the
  /// GeoJSON data has to be set to the [MapShapeLayerDelegate.shapeFile].
  ///
  /// The [MapShapeLayerDelegate.shapeDataField] property is used to
  /// refer the unique field name in the .json file to identify each shapes and
  /// map with the respective data in the data source.
  ///
  /// By default, the value specified for the
  /// [MapShapeLayerDelegate.shapeDataField] in the GeoJSON file will be used in
  /// the elements like data labels, tooltip, and legend for their respective
  /// shapes.
  ///
  /// However, it is possible to keep a data source and customize these elements
  /// based on the requirement. The value of the
  /// [MapShapeLayerDelegate.shapeDataField] will be used to map with the
  /// respective data returned in [MapShapeLayerDelegate.primaryValueMapper]
  /// from the data source.
  ///
  /// Once the above mapping is done, you can customize the elements using the
  /// APIs like [MapShapeLayerDelegate.dataLabelMapper],
  /// [MapShapeLayerDelegate.shapeColorMappers],
  /// [MapShapeLayerDelegate.shapeTooltipTextMapper], etc.
  ///
  /// ```dart
  ///   @override
  ///  Widget build(BuildContext context) {
  ///    return
  ///      SfMaps(
  ///        layers: [
  ///          MapShapeLayer(
  ///            delegate: MapShapeLayerDelegate(
  ///                shapeFile: "assets/world_map.json",
  ///                shapeDataField: "name",
  ///                dataCount: data.length,
  ///                primaryValueMapper: (index) {
  ///                  return data[index].country;
  ///                },
  ///                dataLabelMapper: (index) {
  ///                  return data[index].countryCode;
  ///                }),
  ///        )
  ///      ],
  ///    );
  ///  }
  /// ```
  /// See also:
  /// * [MapShapeLayerDelegate.primaryValueMapper], to map the data of the data
  /// source collection with the respective
  /// [MapShapeLayerDelegate.shapeDataField] in .json file.
  /// * [MapShapeLayerDelegate.bubbleSizeMapper], to customize the bubble size.
  /// * [MapShapeLayerDelegate.dataLabelMapper], to customize the
  /// data label's text.
  /// * [MapShapeLayerDelegate.shapeTooltipTextMapper], to customize the
  /// shape tooltip text.
  /// * [MapShapeLayerDelegate.bubbleTooltipTextMapper], to customize the
  /// bubble tooltip text.
  /// * [MapShapeLayerDelegate.shapeColorValueMapper] and
  /// [MapShapeLayerDelegate.shapeColorMappers], to customize the shape colors.
  /// * [MapShapeLayerDelegate.bubbleColorValueMapper] and
  /// [MapShapeLayerDelegate.bubbleColorMappers], to customize the
  /// bubble colors.
  final MapShapeLayerDelegate delegate;

  /// A builder that specifies the widget to display to the user while the
  /// map is still loading.
  ///
  /// ```dart
  /// @override
  /// Widget build(BuildContext context) {
  ///   return Scaffold(c
  ///      body: Padding(
  ///        padding: EdgeInsets.all(15),
  ///        child: SfMaps(
  ///          layers: <MapLayer>[
  ///            MapShapeLayer(
  ///              delegate: MapShapeLayerDelegate(
  ///                shapeFile: 'assets/world_map.json',
  ///                shapeDataField: 'continent',
  ///              ),
  ///              loadingBuilder: (_) {
  ///                return Container(
  ///                  height: 25,
  ///                  width: 25,
  ///                  child: const CircularProgressIndicator(
  ///                    strokeWidth: 3,
  ///                  ),
  ///                );
  ///              },
  ///            ),
  ///          ],
  ///        ),
  ///      ),
  ///   );
  /// }
  /// ```
  final MapLoadingBuilder loadingBuilder;

  /// Provides option for adding, removing, deleting and updating marker
  /// collection.
  ///
  /// You can also get the current markers count and selected shape's index from
  /// this.
  ///
  /// ```dart
  /// List<Model> data;
  /// MapShapeLayerController controller;
  /// Random random = Random();
  ///
  /// @override
  /// void initState() {
  ///     data = <Model>[
  ///       Model(-14.235004, -51.92528),
  ///       Model(51.16569, 10.451526),
  ///       Model(-25.274398, 133.775136),
  ///       Model(20.593684, 78.96288),
  ///       Model(61.52401, 105.318756)
  ///     ];
  ///
  ///    controller = MapShapeLayerController();
  ///    super.initState();
  /// }
  ///
  /// @override
  /// Widget build(BuildContext context) {
  ///    return Scaffold(
  ///      body: Center(
  ///          child: Container(
  ///            height: 350,
  ///            child: Padding(
  ///              padding: EdgeInsets.only(left: 15, right: 15),
  ///              child: Column(
  ///                children: [
  ///                  SfMaps(
  ///                    layers: <MapLayer>[
  ///                      MapShapeLayer(
  ///                       delegate: MapShapeLayerDelegate(
  ///                          shapeFile: 'assets/world_map.json',
  ///                          shapeDataField: 'name',
  ///                        ),
  ///                        initialMarkersCount: 5,
  ///                        markerBuilder: (_, int index){
  ///                          return MapMarker(
  ///                            latitude: data[index].latitude,
  ///                            longitude: data[index].longitude,
  ///                            child: Icon(Icons.add_location),
  ///                          );
  ///                        },
  ///                        controller: controller,
  ///                      ),
  ///                    ],
  ///                  ),
  ///                  RaisedButton(
  ///                    child: Text('Add marker'),
  ///                    onPressed: () {
  ///                      data.add(Model(
  ///                          -180 + random.nextInt(360).toDouble(),
  ///                          -55 + random.nextInt(139).toDouble()));
  ///                      controller.insertMarker(5);
  ///                    },
  ///                  ),
  ///                ],
  ///              ),
  ///            ),
  ///         )
  ///      ),
  ///   );
  /// }
  ///
  /// class Model {
  ///  Model(this.latitude, this.longitude);
  ///
  ///  final double latitude;
  ///  final double longitude;
  /// }
  /// ```
  final MapShapeLayerController controller;

  /// Enables or disables tooltip for the shapes.
  ///
  /// Defaults to `false`.
  ///
  /// See also:
  /// * [MapTooltipSettings], to customize the tooltip.
  /// * [MapShapeLayerDelegate.shapeTooltipTextMapper], for customizing the
  /// tooltip text.
  final bool enableShapeTooltip;

  /// Enables or disables tooltip for the bubbles.
  ///
  /// Defaults to false.
  ///
  /// See also:
  /// * [MapTooltipSettings], to customize the tooltip.
  /// * [MapShapeLayerDelegate.bubbleTooltipTextMapper], for customizing the
  /// tooltip text.
  final bool enableBubbleTooltip;

  /// Shows or hides the legend.
  ///
  /// Defaults to `false`.
  ///
  /// Information provided in legend items helps to identify the
  /// corresponding data rendered in maps.
  ///
  /// The icons color of the legend is applied based on the colors returned from
  /// the [MapShapeLayerDelegate.shapeColorValueMapper] property and the text is
  /// taken from the [MapShapeLayerDelegate.shapeDataField]. It is possible to
  /// customize the legend icons color and texts using the
  /// [MapColorMapper.color] and [MapColorMapper.text] properties based on the
  /// [MapColorMapper.value] or [MapColorMapper.from] and [MapColorMapper.to]
  /// properties.
  ///
  /// See also:
  /// * [legendSettings], to customize the legend items.
  final bool showLegend;

  /// Shows or hides the data labels.
  ///
  /// Defaults to `false`.
  ///
  /// See also:
  /// * [MapDataLabelSettings], to customize the tooltip.
  /// * [MapShapeLayerDelegate.dataLabelMapper], for customizing the
  /// data label's text.
  final bool showDataLabels;

  /// Shows or hides the bubbles.
  ///
  /// Defaults to `false`.
  ///
  /// See also:
  /// * [MapBubbleSettings], to customize the appearance of the bubbles.
  /// * [MapShapeLayerDelegate.bubbleSizeMapper],
  /// [MapShapeLayerDelegate.bubbleColorMappers],
  /// [MapShapeLayerDelegate.bubbleColorValueMapper] for customizing the
  /// bubbles based on the data.
  final bool showBubbles;

  /// Allows selecting a shape in maps.
  ///
  /// Defaults to `false`.
  ///
  /// See also:
  /// * [MapShapeLayer.onSelectionChanged], for actions to be done during the
  /// shape's selection change.
  /// * [MapSelectionSettings], to customize the selected shape's appearance.
  final bool enableSelection;

  /// Color which is used to paint the shapes.
  final Color color;

  /// Color which is used to paint the stroke of the shapes.
  final Color strokeColor;

  /// Sets the stroke width of the shapes.
  final double strokeWidth;

  /// Paints the shape based on this list of color in a sequential order.
  ///
  /// If the number of shapes exceeds the length of this collections, once the
  /// last color is used for a shape, color in the 0th index will be used for
  /// the next shape and so on.
  ///
  /// ```dart
  /// @override
  /// Widget build(BuildContext context) {
  ///    return Scaffold(
  ///      body: Padding(
  ///        padding: EdgeInsets.all(15),
  ///        child: SfMaps(
  ///          layers: <MapLayer>[
  ///            MapShapeLayer(
  ///              delegate: MapShapeLayerDelegate(
  ///                shapeFile: 'assets/world_map.json',
  ///                shapeDataField: 'continent',
  ///              ),
  ///              palette: [
  ///                Colors.blue[200],
  ///                Colors.orange[200],
  ///                Colors.red[200],
  ///                Colors.green[200],
  ///                Colors.purple[200],
  ///                Colors.lime[200]
  ///              ],
  ///            ),
  ///          ],
  ///        ),
  ///      ),
  ///   );
  /// }
  /// ```
  final List<Color> palette;

  /// Customizes the appearance of the data labels.
  final MapDataLabelSettings dataLabelSettings;

  /// Customizes the appearance of the bubbles.
  ///
  /// See also:
  /// * [MapShapeLayer.showBubbles], to show the bubbles.
  final MapBubbleSettings bubbleSettings;

  /// Customizes the appearance of the the legend.
  ///
  /// See also:
  /// * [MapShapeLayer.showLegend] to render the legend items.
  final MapLegendSettings legendSettings;

  /// Customizes the appearance of the selected shape.
  ///
  /// See also:
  /// * [MapShapeLayer.enableSelection], for allowing shape selection.
  final MapSelectionSettings selectionSettings;

  /// Customizes the bubble and shape tooltip's appearance.
  ///
  /// See also:
  /// * [MapShapeLayer.enableBubbleTooltip], for enabling tooltip for the
  /// bubbles.
  /// * [MapShapeLayer.enableShapeTooltip], for enabling tooltip for the
  /// shapes.
  final MapTooltipSettings tooltipSettings;

  /// Option to select a shape initially.
  ///
  /// See also:
  /// * [enableSelection], to enable shape's selection.
  /// * [MapSelectionSettings], to customize the selected shape's appearance.
  /// * [MapShapeLayer.onSelectionChanged], for actions to be done during the
  /// shape's selection change.
  final int initialSelectedIndex;

  /// Called when the user is selecting a shape by tapping or clicking.
  ///
  /// It passes the index of the selected shape. If the selected shape is
  /// tapped or clicked again, the index will be passed as -1. It means that,
  /// the shape is unselected.
  ///
  /// This snippet shows how to use onSelectionChanged callback in [SfMaps].
  ///
  /// ```dart
  /// SfMaps(
  ///   layers: [MultiChildMapShapeLayer(
  ///       delegate: delegate,
  ///       onSelectionChanged: (int index) {
  ///           print('The selected region is ${data[index].labelText}');
  ///         },
  ///       )]
  /// )
  ///
  /// ```
  final ValueChanged<int> onSelectionChanged;

  @override
  Widget build(BuildContext context) {
    return _MapsShapeLayer(
      key: key,
      delegate: delegate,
      loadingBuilder: loadingBuilder,
      controller: controller,
      initialMarkersCount: initialMarkersCount,
      markerBuilder: markerBuilder,
      showLegend: showLegend,
      showDataLabels: showDataLabels,
      showBubbles: showBubbles,
      enableShapeTooltip: enableShapeTooltip,
      enableBubbleTooltip: enableBubbleTooltip,
      enableSelection: enableSelection,
      color: color,
      strokeColor: strokeColor,
      strokeWidth: strokeWidth,
      palette: palette,
      legendSettings: legendSettings,
      dataLabelSettings: dataLabelSettings,
      bubbleSettings: bubbleSettings,
      selectionSettings: selectionSettings,
      tooltipSettings: tooltipSettings,
      initialSelectedIndex: initialSelectedIndex,
      onSelectionChanged: onSelectionChanged,
    );
  }
}

class _MapsShapeLayer extends StatefulWidget {
  const _MapsShapeLayer({
    Key key,
    this.delegate,
    this.loadingBuilder,
    this.controller,
    this.initialMarkersCount = 0,
    this.markerBuilder,
    this.enableShapeTooltip = false,
    this.enableBubbleTooltip = false,
    this.showLegend = false,
    this.enableSelection = false,
    this.showDataLabels = false,
    this.showBubbles = false,
    this.color,
    this.strokeColor,
    this.strokeWidth,
    this.palette,
    this.legendSettings,
    this.dataLabelSettings,
    this.bubbleSettings,
    this.selectionSettings,
    this.tooltipSettings,
    this.initialSelectedIndex = -1,
    this.onSelectionChanged,
  }) : super(key: key);

  /// Contains the path of the layer source data.
  final MapShapeLayerDelegate delegate;

  final MapLoadingBuilder loadingBuilder;

  /// Coordinates between [SfMaps] and the widget which listens to it.
  final MapShapeLayerController controller;

  /// Sets initial markers count.
  ///
  /// See also:
  /// * [markerBuilder], which returns list of map markers.
  final int initialMarkersCount;

  /// Returns a marker for the given index.
  ///
  /// See also:
  /// * [MapMarker] to create a map marker.
  final MapMarkerBuilder markerBuilder;

  /// Option to show tooltip for the shapes.
  ///
  /// See also:
  /// * [MapTooltipSettings] to customize the tooltip style.
  final bool enableShapeTooltip;

  /// Option to show tooltip for the bubbles.
  ///
  /// See also:
  /// * [MapTooltipSettings] to customize the tooltip style.
  final bool enableBubbleTooltip;

  /// Shows or hides the legend.
  ///
  /// Information provided in legend items helps to identify the
  /// corresponding data rendered in maps.
  ///
  /// If [MapShapeLayerDelegate.colorMappers] is null, then the colors returned
  /// from [MapShapeLayerDelegate.shapeColorValueMapper] will be used for the
  /// legend icon and the legend text will be taken from the
  /// [MapShapeLayerDelegate.primaryValueMapper].
  ///
  /// If [MapShapeLayerDelegate.colorMappers] is not null, then
  /// [MapColorMapper.color] and [MapColorMapper.text] will be used for
  /// legend icon and the legend text respectively.
  ///
  /// See also:
  /// * [legendSettings], to customize the legend items.
  final bool showLegend;

  /// Option to render data labels.
  ///
  /// By default it is false.
  final bool showDataLabels;

  /// Option to display bubbles.
  ///
  /// By default is false.
  final bool showBubbles;

  /// Allows to select a shape in maps.
  ///
  /// By default it is `false`.
  ///
  /// See also:
  /// * [MapSelectionSettings] to customize the selected shape fill,
  /// border colors.
  final bool enableSelection;

  /// Color which is used to paint inside shapes and the edges of shapes.
  ///
  /// This snippet shows how to set fillColor for layer in [SfMaps].
  ///
  /// ```dart
  /// SfMaps(
  ///   layers: [MultiChildMapShapeLayer(
  ///       mapDelegate: mapDelegate,
  ///       layerSettings: MapLayerSettings(
  ///          fillColor: Colors.redAccent.withOpacity(0.5),
  ///          borderColor: Colors.white,
  ///          borderWidth: 0.1)
  ///       )]
  /// )
  ///
  /// ```
  /// See also:
  /// * [strokeColor] paints shape edges of shapes.
  /// * [strokeWidth] draws shape edges based on this value.
  final Color color;

  /// Color which is used to paint edges of shapes.
  final Color strokeColor;

  /// Sets edges width should be drawn.
  final double strokeWidth;

  /// Paints the shape based on this list of color in a sequential order.
  ///
  /// User can set list of colors.
  ///
  /// This snippet shows how to set palette for layer in [SfMaps].
  ///
  /// ```dart
  /// SfMaps(
  ///   layers: [MultiChildMapShapeLayer(
  ///       mapDelegate: mapDelegate,
  ///       layerSettings: MapLayerSettings(
  ///          fillColor: Colors.redAccent.withOpacity(0.5),
  ///          borderColor: Colors.white,
  ///          palette: <Color>[
  ///               Colors.blue[400],
  ///               Colors.blue[300],
  ///               Colors.blue[200],
  ///               Colors.blue[100],
  ///           ],
  ///          borderWidth: 0.1)
  ///       )]
  /// )
  ///
  /// ```
  /// See also:
  /// * [shapeColorValueMapper] signature for getting layer data path color.
  final List<Color> palette;

  /// Option to configure data labels.
  final MapDataLabelSettings dataLabelSettings;

  /// Option to customize the legend icons and labels.
  ///
  /// See also:
  /// * [showLegend] to render the legend items.
  final MapLegendSettings legendSettings;

  /// Option to configure bubble shapes.
  ///
  /// See also:
  /// * [MapShapeLayerDelegate.showBubbles] to render the  bubbles.
  final MapBubbleSettings bubbleSettings;

  final MapSelectionSettings selectionSettings;

  final MapTooltipSettings tooltipSettings;

  /// Option to select a shape initially.It will be called only once.
  ///
  /// See also:
  /// * [MapSelectionSettings] to customize the selected shape fill,
  /// border colors.
  final int initialSelectedIndex;

  /// Called when the user is selecting a shape by tapping.
  ///
  /// It passes the index of the user data.
  ///
  /// This snippet shows how to set onSelectionChanged callback in [SfMaps].
  ///
  /// ```dart
  /// SfMaps(
  ///   layers: [MultiChildMapShapeLayer(
  ///       mapDelegate: mapDelegate,
  ///       onSelectionChanged: (int index) {
  ///           print('The selected region is ${data[index].labelText}');
  ///         },
  ///       )]
  /// )
  ///
  /// ```
  final ValueChanged<int> onSelectionChanged;

  @override
  _MapsShapeLayerState createState() => _MapsShapeLayerState();
}

class _MapsShapeLayerState extends State<_MapsShapeLayer>
    with TickerProviderStateMixin {
  bool isShapeFileDecoded = false;

  bool shouldUpdateMapDataSource = true;

  List<Widget> markers;

  MapLegendSettings legendSettings;

  _MapLegendController legendController;

  AnimationController bubbleAnimationController;

  AnimationController dataLabelAnimationController;

  AnimationController tooltipAnimationController;

  SfMapsThemeData mapsThemeData;

  Timer tooltipDelayTimer;

  Map<String, dynamic> shapeJsonData;

  Map<String, _MapModel> mapDataSource;

  _ShapeBounds shapeBounds;

  double minBubbleValue;

  double maxBubbleValue;

  int initialSelectedIndex;

  int paletteLength;

  @override
  void initState() {
    super.initState();
    assert(widget.delegate != null);
    mapDataSource = <String, _MapModel>{};
    shapeBounds = _ShapeBounds.empty();
    legendController = _MapLegendController();
    dataLabelAnimationController = AnimationController(
        vsync: this, duration: const Duration(milliseconds: 750));
    bubbleAnimationController = AnimationController(
        vsync: this, duration: const Duration(milliseconds: 500));
    tooltipAnimationController = AnimationController(
        vsync: this, duration: const Duration(milliseconds: 200));

    if (widget.controller != null) {
      widget.controller._markersCount = widget.initialMarkersCount;
      widget.controller._selectedIndex = widget.initialSelectedIndex;
    }

    initialSelectedIndex = widget.initialSelectedIndex;
    paletteLength = widget.palette != null ? widget.palette.length : 0;

    MapMarker marker;
    markers = <Widget>[];
    for (int i = 0; i < widget.initialMarkersCount; i++) {
      marker = widget.markerBuilder(context, i);
      assert(marker != null);
      markers.add(marker);
    }

    widget.controller?.addListener(refreshMarkers);
  }

  @override
  void dispose() {
    legendController?.dispose();
    dataLabelAnimationController?.dispose();
    bubbleAnimationController?.dispose();
    tooltipAnimationController?.dispose();

    tooltipDelayTimer?.cancel();

    markers?.clear();
    widget.controller?.removeListener(refreshMarkers);

    mapDataSource?.clear();
    super.dispose();
  }

  @override
  void didUpdateWidget(_MapsShapeLayer oldWidget) {
    assert(widget.delegate != null);
    shouldUpdateMapDataSource = oldWidget.delegate != widget.delegate;
    isShapeFileDecoded =
        oldWidget.delegate.shapeFile == widget.delegate.shapeFile;
    paletteLength = widget.palette != null ? widget.palette.length : 0;
    super.didUpdateWidget(oldWidget);
  }

  @override
  Widget build(BuildContext context) {
    assert(!widget.showDataLabels ||
        (widget.showDataLabels && widget.delegate.shapeDataField != null));
    assert(!widget.showBubbles ||
        widget.showBubbles &&
            widget.delegate.primaryValueMapper != null &&
            widget.delegate.bubbleSizeMapper != null);
    assert(widget.delegate.dataLabelMapper == null ||
        (widget.delegate.dataLabelMapper != null && widget.showDataLabels));
    assert(widget.delegate.bubbleSizeMapper == null ||
        (widget.delegate.bubbleSizeMapper != null && widget.showBubbles));
    assert(widget.delegate.shapeColorMappers == null ||
        widget.delegate.shapeColorMappers.isNotEmpty);
    mapsThemeData = SfMapsTheme.of(context);
    mapsThemeData = mapsThemeData.copyWith(
      layerColor: widget.color ?? mapsThemeData.layerColor,
      layerStrokeColor: widget.strokeColor ?? mapsThemeData.layerStrokeColor,
      layerStrokeWidth: widget.strokeWidth ?? mapsThemeData.layerStrokeWidth,
      legendTextStyle: widget.legendSettings.textStyle ??
          mapsThemeData.legendTextStyle ??
          Theme.of(context).textTheme.caption.copyWith(
              color:
                  Theme.of(context).textTheme.caption.color.withOpacity(0.87)),
      markerIconColor: mapsThemeData.markerIconColor,
      markerIconStrokeColor: mapsThemeData.markerIconStrokeColor,
      markerIconStrokeWidth: mapsThemeData.markerIconStrokeWidth,
      bubbleColor: widget.bubbleSettings.color ?? mapsThemeData.bubbleColor,
      bubbleStrokeColor:
          widget.bubbleSettings.strokeColor ?? mapsThemeData.bubbleStrokeColor,
      bubbleStrokeWidth:
          widget.bubbleSettings.strokeWidth ?? mapsThemeData.bubbleStrokeWidth,
      selectionColor:
          widget.selectionSettings.color ?? mapsThemeData.selectionColor,
      selectionStrokeColor: widget.selectionSettings.strokeColor ??
          mapsThemeData.selectionStrokeColor,
      selectionStrokeWidth: widget.selectionSettings.strokeWidth ??
          mapsThemeData.selectionStrokeWidth,
      tooltipTextStyle: widget.tooltipSettings.textStyle ??
          mapsThemeData.tooltipTextStyle ??
          Theme.of(context)
              .textTheme
              .caption
              .copyWith(color: Theme.of(context).colorScheme.surface),
      tooltipColor: widget.tooltipSettings.color ?? mapsThemeData.tooltipColor,
      tooltipStrokeColor: widget.tooltipSettings.strokeColor ??
          mapsThemeData.tooltipStrokeColor,
      tooltipStrokeWidth: widget.tooltipSettings.strokeWidth ??
          mapsThemeData.tooltipStrokeWidth,
      toggledShapeColor: widget.legendSettings.toggledShapeColor ??
          mapsThemeData.toggledShapeColor,
      toggledShapeStrokeColor: widget.legendSettings.toggledShapeStrokeColor ??
          mapsThemeData.toggledShapeStrokeColor,
      toggledShapeStrokeWidth: widget.legendSettings.toggledShapeStrokeWidth ??
          mapsThemeData.toggledShapeStrokeWidth,
    );

    legendSettings = widget.legendSettings._copyWith(
        textStyle: mapsThemeData.legendTextStyle,
        toggledShapeColor: mapsThemeData.toggledShapeColor,
        toggledShapeStrokeColor: mapsThemeData.toggledShapeStrokeColor,
        toggledShapeStrokeWidth: mapsThemeData.toggledShapeStrokeWidth);

    return FutureBuilder<Map<String, dynamic>>(
      future: _retrieveDataFromShapeFile(
          widget.delegate.shapeFile, shapeJsonData, isShapeFileDecoded),
      builder:
          (BuildContext context, AsyncSnapshot<Map<String, dynamic>> snapshot) {
        if (snapshot.hasData && isShapeFileDecoded) {
          if (shouldUpdateMapDataSource) {
            shapeJsonData = snapshot.data;
            _readJsonFile(snapshot.data);
            minBubbleValue = null;
            maxBubbleValue = null;
            _bindMapsSourceIntoDataSource();
            shouldUpdateMapDataSource = !shouldUpdateMapDataSource;
          }
          return _actualChild(_shapeLayerRenderObjectWidget());
        } else {
          isShapeFileDecoded = true;
          return LayoutBuilder(
            builder: (BuildContext context, BoxConstraints constraints) {
              final Size size = _getActualSize(constraints);
              return Container(
                width: size.width,
                height: size.height,
                alignment: Alignment.center,
                child: widget.loadingBuilder?.call(context),
              );
            },
          );
        }
      },
    );
  }

  Widget _shapeLayerRenderObjectWidget() {
    List<Widget> children = <Widget>[];
    if (markers != null && markers.isNotEmpty) {
      children = List<Widget>.from(markers);
    }

    if (widget.showDataLabels) {
      children.add(_MapDataLabel(
        dataLabelSettings: widget.dataLabelSettings,
        defautTextStyle: Theme.of(context).textTheme.caption,
        state: this,
      ));
    }

    if (widget.showBubbles) {
      children.add(_MapBubble(
          bubbleSettings: widget.bubbleSettings._copyWith(
              color: mapsThemeData.bubbleColor,
              strokeColor: mapsThemeData.bubbleStrokeColor,
              strokeWidth: mapsThemeData.bubbleStrokeWidth),
          state: this));
    }

    if (widget.enableBubbleTooltip || widget.enableShapeTooltip) {
      children.add(_MapTooltip(
        tooltipSettings: widget.tooltipSettings,
        mapDelegate: widget.delegate,
        enableShapeTooltip: widget.enableShapeTooltip,
        enableBubbleTooltip: widget.enableBubbleTooltip,
        mapsThemeData: mapsThemeData,
        state: this,
      ));
    }

    return _MapShapeLayerRenderObjectWidget(
      key: widget.key,
      mapDataSource: mapDataSource,
      mapDelegate: widget.delegate,
      children: children,
      color: mapsThemeData.layerColor,
      strokeColor: mapsThemeData.layerStrokeColor,
      strokeWidth: mapsThemeData.layerStrokeWidth,
      bubbleSettings: widget.bubbleSettings._copyWith(
          color: mapsThemeData.bubbleColor,
          strokeColor: mapsThemeData.bubbleStrokeColor,
          strokeWidth: mapsThemeData.bubbleStrokeWidth),
      enableShapeTooltip: widget.enableShapeTooltip,
      enableBubbleTooltip: widget.enableBubbleTooltip,
      enableSelection: widget.enableSelection,
      legendSettings: legendSettings,
      initialSelectedIndex: widget.initialSelectedIndex,
      selectionSettings: widget.selectionSettings,
      mapsThemeData: mapsThemeData,
      state: this,
    );
  }

  Widget _actualChild(Widget container) {
    if (widget.showLegend) {
      if (legendSettings.offset == null) {
        final MapLegendPosition position = legendSettings.position;
        if (position == MapLegendPosition.top) {
          return Column(
            children: <Widget>[
              _getMapLegend(),
              _getExpandedShapeLayer(container),
            ],
          );
        } else if (position == MapLegendPosition.bottom) {
          return Column(
            children: <Widget>[
              _getExpandedShapeLayer(container),
              _getMapLegend(),
            ],
          );
        } else if (position == MapLegendPosition.left) {
          return Row(
            children: <Widget>[
              _getMapLegend(),
              _getExpandedShapeLayer(container),
            ],
          );
        } else if (position == MapLegendPosition.right) {
          return Row(
            children: <Widget>[
              _getExpandedShapeLayer(container),
              _getMapLegend(),
            ],
          );
        }
      } else {
        return _stackedLegendAndMap(container);
      }
    }

    return container;
  }

  _MapLegend _getMapLegend() {
    return _MapLegend(
        widget.delegate.shapeColorMappers ?? mapDataSource,
        legendSettings,
        mapsThemeData.layerColor,
        legendController,
        mapsThemeData);
  }

  Widget _getExpandedShapeLayer(Widget container) {
    return Expanded(
      child: Stack(
        alignment: Alignment.center,
        children: <Widget>[container],
      ),
    );
  }

  /// Returns the legend and map overlapping widget.
  Widget _stackedLegendAndMap(Widget container) {
    return Stack(
      children: <Widget>[
        _getExpandedShapeLayer(container),
        Align(
          alignment: _getActualLegendAlignment(legendSettings.position),
          // Padding widget is used to set the custom position to the legend.
          child: Padding(
            padding: _getActualLegendOffset(context),
            child: _getMapLegend(),
          ),
        ),
      ],
    );
  }

  /// Returns the alignment for the legend if we set the legend offset.
  AlignmentGeometry _getActualLegendAlignment(MapLegendPosition position) {
    switch (position) {
      case MapLegendPosition.top:
        return Alignment.topCenter;
        break;
      case MapLegendPosition.bottom:
        return Alignment.bottomCenter;
        break;
      case MapLegendPosition.left:
        return Alignment.centerLeft;
        break;
      case MapLegendPosition.right:
        return Alignment.centerRight;
        break;
    }
    return Alignment.topCenter;
  }

  /// Returns the padding value to render the legend based on offset value.
  EdgeInsetsGeometry _getActualLegendOffset(BuildContext context) {
    final Offset offset = legendSettings.offset;
    final MapLegendPosition legendPosition =
        legendSettings.position ?? MapLegendPosition.top;
    // Here the default alignment is center for all the positions.
    // So need to handle the offset by multiplied it by 2.
    switch (legendPosition) {
      // Returns the insets for the offset if the legend position is top.
      case MapLegendPosition.top:
        return EdgeInsets.only(
            left: offset.dx > 0 ? offset.dx * 2 : 0,
            right: offset.dx < 0 ? offset.dx.abs() * 2 : 0,
            top: offset.dy > 0 ? offset.dy : 0);
        break;
      // Returns the insets for the offset if the legend position is left.
      case MapLegendPosition.left:
        return EdgeInsets.only(
            top: offset.dy > 0 ? offset.dy * 2 : 0,
            bottom: offset.dy < 0 ? offset.dy.abs() * 2 : 0,
            left: offset.dx > 0 ? offset.dx : 0);
        break;
      // Returns the insets for the offset if the legend position is right.
      case MapLegendPosition.right:
        return EdgeInsets.only(
            top: offset.dy > 0 ? offset.dy * 2 : 0,
            bottom: offset.dy < 0 ? offset.dy.abs() * 2 : 0,
            right: offset.dx < 0 ? offset.dx.abs() : 0);
        break;
      // Returns the insets for the offset if the legend position is bottom.
      case MapLegendPosition.bottom:
        return EdgeInsets.only(
            left: offset.dx > 0 ? offset.dx * 2 : 0,
            right: offset.dx < 0 ? offset.dx.abs() * 2 : 0,
            bottom: offset.dy < 0 ? offset.dy.abs() : 0);
        break;
    }
    return EdgeInsets.zero;
  }

  void _readJsonFile(Map<String, dynamic> _shapeJsonData) {
    mapDataSource?.clear();
    shapeBounds = _ShapeBounds.empty();

    List<dynamic> polygonGeometryData;
    int multipolygonGeometryLength;
    Map<String, dynamic> geometry;
    Map<String, dynamic> properties;
    final bool hasFeatures = _shapeJsonData.containsKey('features');
    final bool hasGeometries = _shapeJsonData.containsKey('geometries');
    final String key =
        hasFeatures ? 'features' : hasGeometries ? 'geometries' : null;
    final int jsonLength = key.isEmpty ? 0 : _shapeJsonData[key].length;

    for (int i = 0; i < jsonLength; i++) {
      if (hasFeatures) {
        final dynamic features = _shapeJsonData[key][i];
        geometry = features['geometry'];
        properties = features['properties'];
      } else if (hasGeometries) {
        geometry = _shapeJsonData[key][i];
      }

      if (geometry['type'] == 'Polygon') {
        polygonGeometryData = geometry['coordinates'][0];
        _updateMapDataSource(properties, polygonGeometryData);
      } else {
        multipolygonGeometryLength = geometry['coordinates'].length;
        for (int j = 0; j < multipolygonGeometryLength; j++) {
          polygonGeometryData = geometry['coordinates'][j][0];
          _updateMapDataSource(properties, polygonGeometryData);
        }
      }
    }
  }

  void _updateMapDataSource(
      Map<String, dynamic> properties, List<dynamic> points) {
    final String dataPath =
        properties != null ? properties[widget.delegate.shapeDataField] : null;
    mapDataSource.update(
      dataPath,
      (_MapModel model) {
        model.rawPoints.add(points);
        return model;
      },
      ifAbsent: () {
        final int dataSourceIndex = mapDataSource.length;
        return _MapModel(
          primaryKey: dataPath,
          actualIndex: dataSourceIndex,
          shapeColor: paletteLength > 0
              ? widget.palette[dataSourceIndex % paletteLength]
              : null,
          rawPoints: <List<dynamic>>[points],
        );
      },
    );

    _updateShapeBounds(points);
  }

  void _updateShapeBounds(List<dynamic> coordinates) {
    List<dynamic> data;
    num longitude, latitude;
    final int length = coordinates.length;
    for (int i = 0; i < length; i++) {
      data = coordinates[i];
      longitude = data[0];
      latitude = data[1];
      if (shapeBounds.minLongitude == null) {
        shapeBounds.minLongitude = longitude;
        shapeBounds.minLatitude = latitude;
        shapeBounds.maxLongitude = longitude;
        shapeBounds.maxLatitude = latitude;
      } else {
        shapeBounds.minLongitude = min(longitude, shapeBounds.minLongitude);
        shapeBounds.minLatitude = min(latitude, shapeBounds.minLatitude);
        shapeBounds.maxLongitude = max(longitude, shapeBounds.maxLongitude);
        shapeBounds.maxLatitude = max(latitude, shapeBounds.maxLatitude);
      }
    }
  }

  /// Updating [modelSource] data index based on [dataMapper]
  /// value and data color based on [colorValueMapper] value.
  void _bindMapsSourceIntoDataSource() {
    if (widget.delegate.dataCount != null &&
        widget.delegate.dataCount > 0 &&
        widget.delegate.primaryValueMapper != null) {
      final bool hasShapeColorValueMapper =
          widget.delegate.shapeColorValueMapper != null;
      final bool hasDataLabelMapper = widget.delegate.dataLabelMapper != null;
      final bool hasBubbleColorValueMapper =
          widget.delegate.bubbleColorValueMapper != null;
      final bool hasBubbleSizeMapper = widget.delegate.bubbleSizeMapper != null;

      for (int i = 0; i < widget.delegate.dataCount; i++) {
        final _MapModel mapModel =
            mapDataSource[widget.delegate.primaryValueMapper(i)];
        if (mapModel != null) {
          mapModel.dataIndex = i;
          if (hasShapeColorValueMapper) {
            mapModel.shapeColorValue = widget.delegate.shapeColorValueMapper(i);
            mapModel.shapeColor = _getActualColor(
                mapModel.shapeColorValue ?? mapModel.shapeColor,
                widget.delegate.shapeColorMappers,
                mapModel: mapModel);
          }

          if (hasDataLabelMapper) {
            mapModel.dataLabelText = widget.delegate.dataLabelMapper(i);
          }

          if (hasBubbleColorValueMapper) {
            mapModel.bubbleColor = _getActualColor(
                    widget.delegate.bubbleColorValueMapper(i) ??
                        mapModel.bubbleColor,
                    widget.delegate.bubbleColorMappers)
                .withOpacity(widget.bubbleSettings.opacity);
          }

          if (hasBubbleSizeMapper) {
            mapModel.bubbleSizeValue = widget.delegate.bubbleSizeMapper(i);
            if (mapModel.bubbleSizeValue != null) {
              if (minBubbleValue == null) {
                minBubbleValue = mapModel.bubbleSizeValue;
                maxBubbleValue = mapModel.bubbleSizeValue;
              } else {
                minBubbleValue = min(mapModel.bubbleSizeValue, minBubbleValue);
                maxBubbleValue = max(mapModel.bubbleSizeValue, maxBubbleValue);
              }
            }
          }

          if (widget.enableSelection &&
              mapModel.dataIndex == initialSelectedIndex) {
            mapModel.isSelected = true;
            initialSelectedIndex = -1;
          }
        }
      }
    }
  }

  /// Returns color from [MapColorMapper] based on the data source value.
  Color _getActualColor(dynamic colorValue, List<MapColorMapper> colorMappers,
      {_MapModel mapModel}) {
    MapColorMapper mapper;
    final int length = colorMappers != null ? colorMappers.length : 0;
    // Handles equal color mapping.
    if (colorValue is String) {
      for (int i = 0; i < length; i++) {
        mapper = colorMappers[i];
        assert(mapper.value != null && mapper.color != null);
        if (mapper.value == colorValue) {
          mapModel?.actualIndex = i;
          return mapper.color;
        }
      }
    }

    // Handles range color mapping.
    if (colorValue is num) {
      for (int i = 0; i < length; i++) {
        mapper = colorMappers[i];
        assert(
            mapper.from != null && mapper.to != null && mapper.color != null);
        if (mapper.from <= colorValue && mapper.to >= colorValue) {
          mapModel?.actualIndex = i;
          if (mapper.minOpacity != null && mapper.maxOpacity != null) {
            return mapper.color.withOpacity(lerpDouble(
                mapper.minOpacity,
                mapper.maxOpacity,
                (colorValue - mapper.from) / (mapper.to - mapper.from)));
          }
          return mapper.color;
        }
      }
    }

    return colorValue;
  }

  void refreshMarkers() {
    MapMarker marker;
    switch (widget.controller._markerAction) {
      case _MarkerAction.insert:
        marker = widget.markerBuilder(context, widget.controller._index);
        if (widget.controller._index < widget.controller._markersCount) {
          markers.insert(widget.controller._index, marker);
        } else if (widget.controller._index ==
            widget.controller._markersCount) {
          markers.add(marker);
        }
        widget.controller._markersCount++;
        break;
      case _MarkerAction.removeAt:
        assert(widget.controller._index < widget.controller._markersCount);
        markers.removeAt(widget.controller._index);
        widget.controller._markersCount--;
        break;
      case _MarkerAction.replace:
        for (final int index in widget.controller._replaceableIndices) {
          assert(index < widget.controller._markersCount);
          marker = widget.markerBuilder(context, index);
          assert(marker != null);
          markers[index] = marker;
        }
        break;
      case _MarkerAction.clear:
        markers.clear();
        widget.controller._markersCount = 0;
        break;
      default:
        break;
    }

    widget.controller._index = -1;

    setState(() {});
  }
}

class _MapShapeLayerRenderObjectWidget extends Stack {
  _MapShapeLayerRenderObjectWidget({
    Key key,
    List<Widget> children,
    this.mapDataSource,
    this.mapDelegate,
    this.color,
    this.strokeColor,
    this.strokeWidth,
    this.bubbleSettings,
    this.enableShapeTooltip,
    this.enableBubbleTooltip,
    this.enableSelection,
    this.legendSettings,
    this.initialSelectedIndex,
    this.selectionSettings,
    this.mapsThemeData,
    this.state,
  }) : super(
          key: key,
          children: children ?? <Widget>[],
        );

  final Map<String, _MapModel> mapDataSource;

  final MapShapeLayerDelegate mapDelegate;

  final Color color;

  final Color strokeColor;

  final double strokeWidth;

  final MapBubbleSettings bubbleSettings;

  final MapSelectionSettings selectionSettings;

  final bool enableShapeTooltip;

  final bool enableBubbleTooltip;

  final bool enableSelection;

  final MapLegendSettings legendSettings;

  final int initialSelectedIndex;

  final SfMapsThemeData mapsThemeData;

  final _MapsShapeLayerState state;

  @override
  RenderStack createRenderObject(BuildContext context) {
    return _RenderShapeLayer(
      mapDataSource: mapDataSource,
      mapDelegate: mapDelegate,
      color: color,
      strokeColor: strokeColor,
      strokeWidth: strokeWidth,
      bubbleSettings: bubbleSettings,
      enableShapeTooltip: enableShapeTooltip,
      enableBubbleTooltip: enableBubbleTooltip,
      enableSelection: enableSelection,
      legendSettings: legendSettings,
      selectionSettings: selectionSettings,
      mapsThemeData: mapsThemeData,
      state: state,
      mediaQueryData: MediaQuery.of(context),
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, _RenderShapeLayer renderObject) {
    renderObject
      ..mapDataSource = mapDataSource
      ..mapDelegate = mapDelegate
      ..color = color
      ..strokeColor = strokeColor
      ..strokeWidth = strokeWidth
      ..bubbleSettings = bubbleSettings
      ..enableShapeTooltip = enableShapeTooltip
      ..enableBubbleTooltip = enableBubbleTooltip
      ..enableSelection = enableSelection
      ..legendSettings = legendSettings
      ..selectionSettings = selectionSettings
      ..mapsThemeData = mapsThemeData
      ..mediaQueryData = MediaQuery.of(context);
  }
}
